/*
 Wireless Gameboy/GBA Controller - Multi-RX
 Version: 1.3
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 3/10/2018
 Last Modified: 15/05/2019
 
 Use any Gameboy / GBA which supports a Gameboy or GBA cartridge such as the Gameboy, MGB, GBC, GBA, GBA SP, 
 DS, DS Lite and GB Micro as a Wireless game controller.
 
 Using V-USB for USB
 WiiMote code provided by the NES2Wii Project & WiiMote Extension Library.
 
 
 Program the ATmega168PB to use a 12MHz crystal
 avrdude -p ATmeg168PB -c usbasp -U lfuse:w:0xEF:m -U hfuse:w:0xDF:m
 
*/

// ATmega168PB Pin map
//
// PB0 Wiimote SCL, GC/N64 Data In, NES/SNES Clock		PC0 SNES Enable			
// PB1 nRF CE														PC1 GC Enable
// PB2 nRF CSN														PC2 WiiMote Enable											PD2 USB D+
// PB3 nRF MOSI													PC3 NES/SNES Latch											PD3 USB D-
// PB4 nRF MISO													PC4 Wiimote SDA, GC/N64 Data Out, NES/SNES Data		PD4 Wiimote SCL, GC Data In, NES/SNES Clock
// PB5 nRF SCK														PC5 Wiimote SCL, GC/N64 Data In, NES/SNES Clock
// PB6 12MHz XTAL
// PB7 12MHz XTAL																														PD7 LED

#define F_CPU 12000000

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include "setup.c"
#include "mirf.c"
#include "defines.h"
#include "wiimote.c"
#include "transmitters.c"

uint8_t modeFound = 0;
uint16_t packetTimeout = 0;
uint8_t kpData = 0;
uint8_t kpDataHigh = 0;
uint16_t volatile kpOutput = 0xFFFF;
uint8_t firstPacket = 1;
uint8_t kpOutputMapYB = 0;
uint8_t kpOutputMapYA = 0;
uint8_t kpOutputMapRtoA = 0;
uint8_t kpOutputMapLtoA = 0;
uint8_t kpOutputMapLtoZ = 0;
uint8_t kpOutputUseJoystick = 0;
uint8_t volatile timedOut = 0;
uint8_t multiTxData[10];

// Gamecube
uint8_t volatile gcInputBuffer[10];
uint8_t kpOutputDpad = 0;
uint8_t kpOutputZButton = 0;
uint8_t kpOutputLRDownAndClick = 0;

uint8_t kpLHeld = 0;
uint8_t kpLHeldCounter = 0;
uint8_t kpRHeld = 0;
uint8_t kpRHeldCounter = 0;

uint8_t gcReply1[3] = {0x09, 0x00, 0x20};
uint8_t gcReply2[10] = {0x00, 0x80, 0x7C, 0x78, 0x80, 0x79, 0x1A, 0x1B, 0x02, 0x02};
uint8_t gcControllerData[8] = {0x00, 0x80, 0x7D, 0x78, 0x80, 0x7A, 0x1A, 0x1B};

// WiiMote Extension button presses
volatile int8_t jx = 0, jy = 0, rx = 0, ry = 0;
volatile uint8_t tl = 0, tr = 0;
volatile uint8_t left = 0, right = 0, up = 0, down = 0, a = 0, b = 0, x = 0, y = 0, 
						select = 0, start = 0, home = 0, l = 0, r = 0, zl = 0, zr = 0;

// N64
uint8_t volatile n64ControllerData[4] = {0x00, 0x00, 0x00, 0x00};


int main (void) {
	setup();
	
	// Init the adapter
	if (adapterConnected == USB_CONNECTED) {
		usbInit(); // Start USB
	}
	else if (adapterConnected == GC_CONNECTED) {
		gcInputBuffer[0] = 0;
	}
	
	check_for_new_settings(); // Check for new nRF settings
	check_for_tx_mode(); // Check for TX modee
	
	// Turn on time out counter for ~5 seconds
	TCNT1 = 0;
	TIMSK1 = (1<<TOV1);
	TCCR1B = (1<<CS12) | (1<<CS10);
	
	// Wait for latch low for NES/SNES
	if (adapterConnected == NES_CONNECTED || adapterConnected == SNES_CONNECTED) {
		while (!(PINC & (1<<NES_SNES_LATCH))); // Wait until high
		_delay_us(500);
		
		PCMSK1 |= (1<<PCINT11); // Turn on pin interrupt on latch pin
		PCICR |= (1<<PCIE1);
	}
	else if (adapterConnected == GC_CONNECTED) { // Wait for data low for GC
		while (PINC & (1<<GC_DATA_IN));
		_delay_us(500);
		
		PCMSK2 |= (1<<PCINT20); // Turn on pin interrupt on data pin
		PCICR |= (1<<PCIE2);
	}
	else if (adapterConnected == N64_CONNECTED) {
		while (PINC & (1<<N64_DATA_IN));
		_delay_us(500);
		
		PCMSK0 |= (1<<PCINT0);// Turn on pin interrupt on data pin
		PCICR |= (1<<PCIE0);
	}
	
	while(1) {
		if (timedOut == 1) { // If timed out, reset to defaults and check for the transmitter again
			// NES/SNES
			kpOutput = 0xFFFF; 
			
			// USB
			reportBuffer.X_JOY = 0;
			reportBuffer.Y_JOY = 0;
			reportBuffer.Z_JOY = 0;
			reportBuffer.RX_JOY = 0;
			reportBuffer.RY_JOY = 0;
			reportBuffer.RZ_JOY = 0;
			reportBuffer.B = 0;
			reportBuffer.A = 0;
			reportBuffer.SELECT = 0;
			reportBuffer.START = 0;
			reportBuffer.L = 0;
			reportBuffer.R = 0;
			reportBuffer.X = 0;
			reportBuffer.Y = 0;
			reportBuffer.Z = 0;
			reportBuffer.D_UP = 0;
			reportBuffer.D_DOWN = 0;
			reportBuffer.D_LEFT = 0;
			reportBuffer.D_RIGHT = 0;
			reportBuffer.T8 = 0;
			reportBuffer.T9 = 0;
			reportBuffer.T10 = 0;
			
			// Wii
			left = 0, right = 0, up = 0, down = 0, a = 0, b = 0, x = 0, y = 0, 
			select = 0, start = 0, home = 0, l = 0, r = 0, zl = 0, zr = 0;
			
			// GC
			gcControllerData[0] = 0;
			gcControllerData[1] = 0x80;
			gcControllerData[2] = 0x7D;
			gcControllerData[3] = 0x78;
			gcControllerData[6] = 0x1A;
			gcControllerData[7] = 0x1B;
			kpLHeld = 0;
			kpRHeld = 0;
			
			// N64
			n64ControllerData[0] = 0;
			n64ControllerData[1] = 0;
			n64ControllerData[2] = 0;
			n64ControllerData[3] = 0;
			
			check_for_tx_mode();
			timedOut = 0;
			TCNT1 = 0;
			TCCR1B = (1<<CS12) | (1<<CS10);
		}
		
		if (mirf_status() & (1<<RX_DR)) { // Packet received
			PORTD |= (1<<LED);
			mirf_CE_lo; // Stop listening
			
			mirf_CSN_lo; // Pull down chip select
			spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
			
			if (modeFound == SNES_TX) { // Read 3 bytes if in SNES mode
				for (uint8_t x = 0; x < 3; x++) {
					multiTxData[x] = spi_transfer(0);
				}
			}
			else if (modeFound == GBA_MODE) { // Load 2 bytes if in GBA mode
				kpDataHigh = spi_transfer(0);
				kpData = spi_transfer(0);
			}
			else {
				kpData = spi_transfer(0); // Read data
			}
			
			mirf_CSN_hi; // Pull up chip select
			mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
			
			mirf_flush_rx_tx(); // Flush TX/RX
			mirf_CE_hi; // Start listening
			PORTD &= ~(1<<LED);
			
			// Reset timeout timer
			TCNT1 = 0;
			
			// Check for key presses
			if (adapterConnected == NES_CONNECTED) {
				nes_adapter();
			}
			else if (adapterConnected == SNES_CONNECTED) {
				snes_adapter();
			}
			else if (adapterConnected == N64_CONNECTED) {
				n64_adapter();
			}
			else if (adapterConnected == GC_CONNECTED) {
				gc_adapter();
			}
			else if (adapterConnected == WII_CONNECTED) {
				wii_adapter();
			}
		}
		
		// USB Poll
		if (adapterConnected == USB_CONNECTED) {
			usb_adapter();
		}
		
		_delay_us(250);
	}
	
	return 0;
}

// NES/SNES - Latch pin was toggled
ISR(PCINT1_vect) {
	if (!(PINC & (1<<NES_SNES_LATCH))) { // Latch low
		PORTD &= ~(1<<LED);
		uint8_t latchReset = 0;
		
		// 8/16 bits of data to write depending on what keys were pressed
		for (uint16_t x = loopTimes; x != 0; x >>= 1) { // Loop 8/16 times
			if (kpOutput & x) {
				PORTC |= (1<<NES_SNES_DATA);
			}
			else {
				PORTC &= ~(1<<NES_SNES_DATA);
			}
			
			while (PINC & (1<<NES_SNES_CLK)) { // Wait for clock low
				if (PINC & (1<<NES_SNES_LATCH)) { // Latch is high, break out of everything
					latchReset = 1;
					break;
				}
			}
			if (latchReset == 1) {
				break;
			}
			while (!(PINC & (1<<NES_SNES_CLK))); // Wait for clock high
		}
		PORTC &= ~(1<<NES_SNES_DATA); // Set data low (required for SNES)
	}
	else { // Latch high, set first bit
		if (kpOutput & loopTimes) {
			PORTC |= (1<<NES_SNES_DATA);
		}
		else {
			PORTC &= ~(1<<NES_SNES_DATA);
		}
	}
	_delay_us(5);
}

// GC - Data input was toggled
ISR(PCINT2_vect) {
	if (!(PINC & (1<<GC_DATA_IN))) {
		PORTD &= ~(1<<LED);
		
		// Read 1 to 3 bytes depending the first byte we receive
		for (uint8_t byte = 0; byte < 3; byte++) {
			uint8_t dataIn = 0;
			for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
				uint8_t bitTime = 0;
				
				while (!(PINC & (1<<GC_DATA_IN))); // Wait until high
				//PORTD |= (1<<LED);
				asm volatile("nop");
				
				while (PINC & (1<<GC_DATA_IN)) { // While high, count
					bitTime++;
				}
				//PORTD &= ~(1<<LED);
				asm volatile("nop");
				
				if (bitTime >= 6) { // High
					dataIn |= x;
				}
			}
			gcInputBuffer[byte] = dataIn;
			
			// Only read the first byte, exit
			if (gcInputBuffer[0] == 0 || gcInputBuffer[0] == 0x41) {
				break;
			}
		}
		
		while (!(PINC & (1<<GC_DATA_IN))); // Wait until high
		asm volatile("nop");
		
		// Reply back depending on what the first byte was
		if (gcInputBuffer[0] == 0) {
			for (uint8_t byte = 0; byte < 3; byte++) {
				for (uint8_t x = 0x80; x != 0; x >>= 1) {
					if (gcReply1[byte] & x) {
						PORTC |= (1<<GC_DATA_OUT);
						asm volatile("nop");
						asm volatile("nop");
						PORTC &= ~(1<<GC_DATA_OUT);
						_delay_us(2);
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						PORTC |= (1<<GC_DATA_OUT);
					}
					else {
						PORTC |= (1<<GC_DATA_OUT);
						_delay_us(2);
						asm volatile("nop");
						asm volatile("nop");
						PORTC &= ~(1<<GC_DATA_OUT);
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						PORTC |= (1<<GC_DATA_OUT);
					}
				}
			}
			PORTC |= (1<<GC_DATA_OUT);
			_delay_us(2);
			PORTC &= ~(1<<GC_DATA_OUT);
		}
		else if (gcInputBuffer[0] == 0x41 || gcInputBuffer[0] == 0x42 || gcInputBuffer[0] == 0x43) {
			for (uint8_t byte = 0; byte < 10; byte++) {
				for (uint8_t x = 0x80; x != 0; x >>= 1) {
					if (gcReply2[byte] & x) {
						PORTC |= (1<<GC_DATA_OUT);
						asm volatile("nop");
						asm volatile("nop");
						PORTC &= ~(1<<GC_DATA_OUT);
						_delay_us(2);
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						PORTC |= (1<<GC_DATA_OUT);
					}
					else {
						PORTC |= (1<<GC_DATA_OUT);
						_delay_us(2);
						asm volatile("nop");
						asm volatile("nop");
						PORTC &= ~(1<<GC_DATA_OUT);
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						asm volatile("nop");
						PORTC |= (1<<GC_DATA_OUT);
					}
				}
			}
			PORTC |= (1<<GC_DATA_OUT);
			_delay_us(2);
			PORTC &= ~(1<<GC_DATA_OUT);
		}
		else if (gcInputBuffer[0] == 0x40) {
			// Write the controller data, needs to be done quickly (without a main loop) or Brook GC to Switch adapter won't work properly
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[0] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[1] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[2] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[3] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[4] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[5] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[6] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			for (uint8_t x = 0x80; x != 0; x >>= 1) {
				if (gcControllerData[7] & x) {
					PORTC |= (1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
				else {
					PORTC |= (1<<GC_DATA_OUT);
					_delay_us(2);
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<GC_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC |= (1<<GC_DATA_OUT);
				}
			}
			
			PORTC |= (1<<GC_DATA_OUT);
			_delay_us(2);
			PORTC &= ~(1<<GC_DATA_OUT);
		}
	}
	
	_delay_us(5); // Wait a little so we don't re-trigger the interrupt
}

// N64 - Data pin was toggled
ISR(PCINT0_vect) {
	if (!(PINC & (1<<N64_DATA_IN))) {
		PORTD &= ~(1<<LED);
		
		uint8_t dataIn = 0;
		for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
			uint8_t bitTime = 0;
			
			while (!(PINC & (1<<N64_DATA_IN))); // Wait until high
			//PORTA |= (1<<LED);
			asm volatile("nop");
			
			while (PINC & (1<<N64_DATA_IN)) { // While high, count
				bitTime++;
			}
			//PORTA &= ~(1<<LED);
			asm volatile("nop");
			
			if (bitTime >= 6) { // High
				dataIn |= x;
			}
		}
		
		while (!(PINC & (1<<N64_DATA_IN))); // Wait until high
		_delay_us(1);
		
		// Write the controller data
		for (uint8_t byte = 0; byte < 4; byte++) {
			for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
				if (n64ControllerData[byte] & x) {
					PORTC |= (1<<N64_DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTC &= ~(1<<N64_DATA_OUT);
					_delay_us(3);
					PORTC |= (1<<N64_DATA_OUT);
				}
				else {
					PORTC |= (1<<N64_DATA_OUT);
					_delay_us(2);
					PORTC &= ~(1<<N64_DATA_OUT);
					_delay_us(1);
					PORTC |= (1<<N64_DATA_OUT);
				}
			}
		}
		PORTC |= (1<<N64_DATA_OUT);
		_delay_us(2);
		PORTC &= ~(1<<N64_DATA_OUT);
		
		_delay_us(5);
	}
}

// Overflows after ~5 seconds of no packets received
ISR(TIMER1_OVF_vect) {
	timedOut = 1;
	TCNT1 = 0;
	TCCR1B = 0;
}
